import json
import os
import time

from google.cloud.monitoring_v3 import (
    MetricServiceClient,
    TimeInterval,
    ListTimeSeriesRequest,
)
from google.cloud import storage

from verinfast.utils.utils import std_exec


def get_bucket_size(bucket_name):
    try:
        cmd = ["gcloud", "storage", "du", f"gs://{bucket_name}", "--summarize"]
        output = std_exec(cmd)
        # Extracting size from the output
        for line in output.splitlines():
            if line.startswith("Total Size:"):
                size_str = line.split(":")[1].strip()
                size_bytes = int(size_str.split()[0])
                return size_bytes
    except Exception as e:
        print(f"Error getting size for bucket {bucket_name}: {e}")
    return 0


def getBlocks(sub_id: str, path_to_output: str = "./", dry=False):
    if not dry:
        # Instantiates a client
        storage_client = storage.Client(project=sub_id)

        # List all the buckets available
        client = MetricServiceClient()
        # for bucket in storage_client.list_buckets():
        #     print(bucket)
        #     print(bucket.name)

        now = time.time()
        seconds = int(now)
        nanos = int((now - seconds) * 10**9)
        interval = TimeInterval(
            {
                "end_time": {"seconds": seconds, "nanos": nanos},
                "start_time": {"seconds": (seconds - 3000), "nanos": nanos},
            }
        )

        results = client.list_time_series(
            request={
                "name": f"projects/{sub_id}",
                "filter": 'metric.type = "storage.googleapis.com/storage/total_bytes"',
                "interval": interval,
                "view": ListTimeSeriesRequest.TimeSeriesView.FULL,
            }
        )
        my_buckets = []
        known_buckets = {}
        all_buckets = storage_client.list_buckets()
        for bucket in all_buckets:
            rp = bucket.retention_period
            size = get_bucket_size(bucket.name)
            known_buckets[bucket.name] = {
                "name": bucket.name,
                "size": size,
                "retention": str(rp),
                "public": False,
            }
            iam = bucket.get_iam_policy()
            permissions = []
            for b in iam.bindings:
                p = {"permission": b["role"], "roles": list(b["members"])}
                permissions.append(json.dumps(p))
                if "allUsers" in b["members"]:
                    known_buckets[bucket.name]["public"] = True
            known_buckets[bucket.name]["permissions"] = permissions

        for result in results:
            if result.resource and result.resource.labels:
                bn = result.resource.labels["bucket_name"]
                size = result.points[-1].value.double_value
                if bn in known_buckets:
                    known_buckets[bn]["size"] = size
        my_buckets = list(known_buckets.values())
        upload = {
            "metadata": {"provider": "gcp", "account": str(sub_id)},
            "data": my_buckets,
        }
    # End dry block
    gcp_output_file = os.path.join(path_to_output, f"gcp-storage-{sub_id}.json")
    if not dry:
        with open(gcp_output_file, "w") as outfile:
            outfile.write(json.dumps(upload, indent=4))
    return gcp_output_file


# Test Code
# getBlocks(sub_id="startupos-328814")



import json

from verinfast.dependencies.walkers.classes import Walker, Entry


class PackageWalker(Walker):
    def remote_decorate(self, entry: Entry) -> str:
        resp = None
        try:
            license_resp = self.getUrl(
                f"https://registry.npmjs.org/{entry.name}/{entry.specifier}/"
            )
            resp = json.loads(license_resp)
        except Exception as e:
            self.log(
                f"License not found for: {entry.name}@{entry.specifier}", display=False
            )
            self.log(e, display=False)
        if isinstance(resp, dict):
            entry.license = resp.get("license", "License not available")
            entry.summary = resp.get("description", "No description provided.")
        else:
            entry.license = "License not available"
            entry.summary = "No description provided."
            self.log(f"Error with {entry.name} {entry.specifier} response")
            self.log(license_resp)

    def parse(self, file: str, expand=False):
        try:
            with open(file) as f:
                d = json.load(f)
                if "dependencies" in d:
                    dependencies = d["dependencies"]
                    for dependency in dependencies:
                        try:
                            e = Entry(
                                name=dependency,
                                specifier=dependencies[dependency].get(
                                    "version", "Version not found"
                                ),
                                source="package-lock.json",
                            )
                            self.remote_decorate(entry=e)
                            self.entries.append(e)
                        except Exception as error:
                            self.log(f"error parsing {dependency}")
                            self.log(error)
                            continue
        except Exception as error:
            # handle the exception
            self.log(f"error parsing {file}")
            self.log(error)



from datetime import date
import inspect
import os
import re
from glob import glob
import subprocess
import time
from typing import List, Union


STD_EXCLUDE_LIST = [
    "**/.git/**",
    "**/node_modules/**",
    "build/**",
    "dist/**",
    "env/**",
    "venv/**",
    "**/__pycache__/**",
    "**/*.pyc",
    "**/*.pyo",
    "**/*.log",
    "**/*.swp",
    "**/*.swo",
    "**/*.DS_Store",
    "**/.idea/**",
]


"""
newline = "\n" #
TODO - Set to system appropriate newline character.
This doesn't work with modernmetric
"""


def list_files(
    path: str, incl: List[glob] = ["*.*"], excl: List[glob] = STD_EXCLUDE_LIST
):
    files = []
    excluded = []
    for i in incl:
        files += glob(os.path.join(path, i))
    for e in excl:
        excluded += glob(os.path.join(path, e))
    files = [f for f in files if f not in excluded]
    return files


def std_exec(cmd: List[str], log=None):
    try:
        return subprocess.check_output(cmd).decode("utf-8")
    except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
        try:
            return subprocess.check_output(cmd, shell=True).decode(encoding="latin-1")
        except Exception as e2:
            if log is not None:
                log(tag="std_exec replace Error", msg=f"{e2}, {cmd}")
            else:
                print(f"std_exec replace Error: {e2}, {cmd}")
            if log is not None:
                log(tag="std_exec Error", msg=f"{e}, {cmd}")
            else:
                print(f"std_exec Error: {e}, {cmd}")
            return ""


def escapeChars(text: str):
    fixedText = re.sub(r"([\"\{\}])", r"\\\1", text)
    return fixedText


def trimLineBreaks(text: str):
    return text.replace("\n", "").replace("\r", "")


# Truncate large strings for display
def truncate(text, length=100):
    testStr = str(text)  # Supports passing in Lists and other types
    return (testStr[:length] + "..") if len(testStr) > length else testStr


def truncate_children(
    obj: Union[dict, list], log, excludes=[], max_length=30, recursion_depth=0
):
    if isinstance(obj, dict):
        for k in obj:
            v = obj[k]
            if k in excludes:
                pass
            elif isinstance(v, str):
                obj[k] = v[0:max_length]
            elif isinstance(v, int) or isinstance(v, float) or isinstance(v, bool):
                pass
            else:
                obj[k] = truncate_children(
                    obj[k],
                    log,
                    excludes=excludes,
                    max_length=max_length,
                    recursion_depth=recursion_depth + 1,
                )
    elif isinstance(obj, list):
        for i, v in enumerate(obj):
            if isinstance(v, str):
                obj[i] = v[0:max_length]
            elif isinstance(v, int) or isinstance(v, float) or isinstance(v, bool):
                pass
            else:
                obj[i] = truncate_children(
                    obj[i],
                    log,
                    excludes=excludes,
                    max_length=max_length,
                    recursion_depth=recursion_depth + 1,
                )

    return obj


# Chunk a list
def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


# DebugLog
# Usage: must pass in full path to a log
# file as "file" OR pass in a path and the
# class will create a default log file called
# "log.txt" in the specified path
class DebugLog:
    def __init__(self, path: str = None, file: str = None, debug: bool = False):
        if path is None and file is None:
            raise ValueError("No log file or path specified")
            return
        if path is not None and file is not None:
            raise ValueError("Both log file and path specified")
            return
        if path is not None and file is None:
            self.path = path
            self.file = os.path.join(path, "log.txt")  # Default log file
        if path is None and file is not None:
            self.file = file
        self.debug = debug

    def log(self, msg, tag=None, display=False, timestamp=True):
        if tag is None and timestamp:
            s = inspect.stack()[1]

            tag = f"{s.filename}@{s.lineno} {s.function}"
        if timestamp:
            d = date.today()
            tag = f"{d} {time.strftime('%H:%M:%S', time.localtime())} {tag}"
        if tag != "":
            output = f"{tag}: {msg}"
        else:
            output = f"{msg}"

        with open(self.file, "a") as f:
            f.write(output + "\n")

        if display or self.debug:
            print(output)


# Returns a dict of the repo name, repo url and branch from the original url
def get_repo_name_url_and_branch(repo_url: str):
    # The regex pattern captures the entire URL
    # and the repository name after the last '/'

    repo_url = repo_url.strip()
    branch = None

    split_url = repo_url.split("@")

    # HTTPS URLs
    if repo_url.startswith("https://") or repo_url.startswith("http://"):
        # Check for branch in URL
        if len(split_url) == 2:
            branch = split_url[1]
            repo_url = split_url[0]
        elif len(split_url) == 3:
            # Username and password in URL
            # e.g. https://user:password@my.githost.com/repo@branch
            branch = split_url[2]
            # Remove branch so that the repo name can be extracted
            # without colliding with /'s in branch name
            repo_url = repo_url.replace(f"@{branch}", "")
    else:
        # SSH URLs
        # Check for branch in URL
        if len(split_url) > 2:
            branch = split_url[2]
            # Remove branch so that the repo name can be extracted
            # without colliding with /'s in branch name
            repo_url = repo_url.replace(f"@{branch}", "")
            print(f"Repo URL after removing branch: {repo_url}")
            print(f"@{branch}")

    # Match repo_name to after the last '/' in the cleaned URL
    repo_name = repo_url.rsplit("/", 1)[-1]

    return {"repo_name": repo_name, "repo_url": repo_url, "branch": branch}



import contextlib
import io
import json
import os
from pathlib import Path
import platform
import time

import semgrep.commands.scan as semgrep_scan
from cachehash.main import Cache

from verinfast.config import Config
from verinfast.utils.utils import truncate_children


def run_scan(
    path: str,
    repo_name: str,
    config: Config,
    cache: Cache,
    upload,
    template_definition: dict,
    log=print,
) -> None:
    path = str(Path(path).absolute())
    log(f"path: {path}")
    uname = platform.uname()
    system = uname.system

    if not config.runScan:
        return

    if system.lower() == "windows":
        log(
            """
        Windows does not support Semgrep.
        Please see the open issues here:
        https://github.com/returntocorp/semgrep/issues/1330
                    """
        )

        return

    findings_file = os.path.join(config.output_dir, f"{repo_name}.findings.json")

    custom_args = ["--config", "auto", "--json", f"--output={findings_file}", "-q"]

    findings_success = False
    if not config.dry:
        # if cache exists write cache to findings file
        start_time = time.time()
        cache_results = cache.get(path)
        first_duration = time.time() - start_time
        log(f"cache get took: {first_duration:.2f} seconds")

        if cache_results:
            log("CACHE HIT")
            with open(findings_file, "w") as f:
                f.write(json.dumps(cache_results))
                findings_success = True

        # else run scan
        else:
            log("cache miss")
            log(msg=repo_name, tag="Scanning repository", display=True)
            try:
                with contextlib.redirect_stdout(io.StringIO()):
                    semgrep_scan.scan(custom_args)
                    findings_success = True
            except SystemExit as e:
                if e.code == 0:
                    findings_success = True
                else:
                    log(tag="ERROR", msg="SystemExit in Semgrep")
                    log(e)
            except Exception as e:
                log(tag="ERROR", msg="Error in Semgrep")
                log(e)

            # set cache
            # Only try to cache if scan was successful and file exists
            if findings_success and os.path.exists(findings_file):
                try:
                    # Try to cache the results
                    with open(findings_file) as f:
                        results = json.load(f)
                        cache.set(path, results)
                except Exception as e:
                    log(tag="Cache Error", msg=f"Failed to cache results: {str(e)}")

        if findings_success:
            try:
                with open(findings_file) as f:
                    findings = json.load(f)

                # This is on purpose. If you try to read same pointer
                # twice, it dies.
                with open(findings_file) as f:
                    original_findings = json.load(f)

                if config.truncate_findings:
                    # Exclusions are set to exclude fields that are not code
                    truncation_exclusion = [
                        "cwe",
                        "owasp",
                        "path",
                        "check_id",
                        "license",
                        "fingerprint",
                        "message",
                        "references",
                        "url",
                        "source",
                        "severity",
                    ]
                    log(tag="TRUNCATING", msg=f"excluding: {truncation_exclusion}")
                    try:
                        findings = truncate_children(
                            findings,
                            log,
                            excludes=truncation_exclusion,
                            max_length=config.truncate_findings_length,
                        )
                    except Exception as e:
                        log(tag="ERROR", msg="Error in Truncation")
                        log(e)
                        log(json.dumps(original_findings, indent=4, sort_keys=True))
                with open(findings_file, "w") as f2:
                    f2.write(json.dumps(findings, indent=4, sort_keys=True))
                template_definition["gitfindings"] = findings
            except Exception as e:
                if not config.dry:
                    log(tag="ERROR", msg="Error in findings post-processing")
                    log(e)
                else:
                    log(
                        msg=f"""
                            Attempted to format/truncate non existent file
                            {findings_file}
                        """
                    )
        else:
            log(msg="Scan Findings failed")

    # End if findings_success is True

    # Upload findings always, in case of dry run
    # .upload checks should_upload
    upload(file=findings_file, route="findings", source=repo_name)



import json
import os
from pathlib import Path
import time

from verinfast.agent import Agent
from verinfast.code_scan import run_scan
from verinfast.config import Config
from verinfast.utils.utils import DebugLog

from cachehash.main import Cache
from unittest.mock import patch

# Create test file
pwd = os.path.dirname(__file__)
test_file = Path(pwd, "fixtures", "bad_password.py")
test_folder = Path(pwd, "fixtures").absolute()


# Minimal configuration
config = Config()
config.runGit = False
config.runSizes = False
config.runStats = False
config.runDependencies = False
config.config["local_repos"] = [str(test_file.parent.absolute())]

# Create Cache
db_path = "semgrep.db"
db_path = Path(pwd, db_path)

test_log_path = Path(pwd, "test_log.txt")
mock_log = DebugLog(file=str(test_log_path), debug=True).log


def setup_database():
    if db_path.exists():
        os.remove(str(db_path))
    assert not db_path.exists()

    if test_log_path.exists():
        os.remove(str(test_log_path))
    assert not test_log_path.exists()


def test_cache_persistence():
    cache_path = Path(Path.home(), ".verinfast_cache/semgrep.db")

    # First run creates cache
    agent = Agent()
    agent.config.runGit = False
    agent.config.runSizes = False
    agent.config.runStats = False
    agent.config.runDependencies = False
    agent.scan()

    # Verify cache exists and has content
    assert cache_path.exists()
    assert cache_path.stat().st_size > 0


@patch("verinfast.user.__get_input__", return_value="y")
def test_semgrep_cache(self):
    setup_database()
    cache = Cache(path=db_path, table="test_cache")

    def mock_upload(*args, **kwargs):
        for k in kwargs:
            if k != "file":
                assert expected_upload[k] == kwargs[k]
            else:
                with open(kwargs[k]) as f:
                    res = json.load(f)
                    r = res["results"]
                    result = r[0]
                    assert result["extra"]["metadata"]["cwe"] is not None

    expected_upload = {"route": "findings", "source": "test"}

    # First run
    print("\nRunning first scan...")
    test_args = {
        "repo_name": "test",
        "path": test_folder,
        "config": config,
        "cache": cache,
        "upload": mock_upload,
        "template_definition": {},
        "log": mock_log,
    }
    start_time = time.time()
    run_scan(**test_args)

    first_duration = time.time() - start_time
    print(f"First scan took: {first_duration:.2f} seconds")

    # Second run should use cache
    print("\nRunning second scan...")
    second_start_time = time.time()
    run_scan(**test_args)
    second_duration = time.time() - second_start_time
    print(f"Second scan took: {second_duration:.2f} seconds")

    # Second run should be faster
    assert second_duration < (first_duration - 5)
    assert db_path.exists()



import datetime
import json
import os

from azure.identity import DefaultAzureCredential
from azure.monitor.query import MetricsQueryClient
from azure.mgmt.storage import StorageManagementClient
from azure.mgmt.resource import ResourceManagementClient


def getBlocks(sub_id: str, path_to_output: str = "./", dry=False):
    if not dry:
        credential = DefaultAzureCredential()
        client = MetricsQueryClient(credential)
        resource_client = ResourceManagementClient(credential, subscription_id=sub_id)

        group_list = resource_client.resource_groups.list()
        storage_client = StorageManagementClient(credential, subscription_id=sub_id)

        known_buckets = {}
        for group in group_list:
            accounts = storage_client.storage_accounts.list_by_resource_group(
                resource_group_name=group.name
            )

            for account in accounts:
                containers = storage_client.blob_containers.list(
                    resource_group_name=group.name, account_name=account.name
                )

                d = datetime.timedelta(
                    days=1,
                    seconds=0,
                    microseconds=0,
                    milliseconds=0,
                    minutes=0,
                    hours=0,
                    weeks=0,
                )

                o = client.query_resource(
                    resource_uri=account.id, metric_names=["UsedCapacity"], timespan=d
                )

                bytes = o.metrics[0].timeseries[0].data[0].average
                print(bytes)
                known_buckets[account.name] = {
                    "name": account.name,
                    "size": bytes,
                    "retention": None,
                    "public": False,
                    "permissions": [],
                }
                for c in containers:
                    if c.public_access:
                        known_buckets[account.name]["public"] = True
        my_buckets = list(known_buckets.values())
        upload = {
            "metadata": {"provider": "azure", "account": str(sub_id)},
            "data": my_buckets,
        }
    # End dry block
    azure_output_file = os.path.join(path_to_output, f"az-storage-{sub_id}.json")
    if not dry:
        with open(azure_output_file, "w") as outfile:
            outfile.write(json.dumps(upload, indent=4))
    return azure_output_file


# Test Code
# getBlocks(sub_id="80dc7a6b-df94-44be-a235-7e7ade335a3c")



import json
import os
from unittest.mock import patch
from pathlib import Path
import shutil

import pytest

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.utils.utils import DebugLog
from typing import Union


file_path = Path(__file__)
test_folder = file_path.parent.absolute()
str_path = str(test_folder.joinpath("str_conf.yaml").absolute())
results_dir = test_folder.joinpath("results").absolute()
saw_message = False
MAX_RECURSION_DEPTH = 10


def check_children(
    i: Union[str, dict, list],
    max_length=30,
    recursion_depth=0,
    # This list must match agent.py
    excludes=[
        "cwe",
        "owasp",
        "path",
        "check_id",
        "license",
        "fingerprint",
        "message",
        "references",
        "url",
        "source",
        "severity",
    ],
):
    if recursion_depth > MAX_RECURSION_DEPTH:
        raise Exception("In TOO DEEP!")
    global saw_message
    # print(i)
    if isinstance(i, str):
        if len(i) > max_length:
            print(i)
        assert len(i) <= max_length
    elif isinstance(i, dict):
        for k in i:
            if k == "message":
                saw_message = True
            if k in excludes:
                print(f"{k} is excluded")
            else:
                try:
                    check_children(
                        i[k], recursion_depth=recursion_depth + 1, excludes=excludes
                    )
                except Exception as e:
                    print(k)
                    raise e
    elif isinstance(i, list):
        for k in i:
            try:
                check_children(
                    k, recursion_depth=recursion_depth + 1, excludes=excludes
                )
            except Exception as e:
                print(k)
                raise e
    elif isinstance(i, float) or isinstance(i, int) or isinstance(i, bool):
        pass
    else:
        raise Exception("Non-serializable Object")


@patch("verinfast.user.__get_input__", return_value="y")
def test_no_truncate(self):
    global saw_message
    saw_message = False
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    config.output_dir = results_dir
    agent.config = config
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.scan()
    findings_file_path = results_dir.joinpath("small-test-repo.git.findings.json")

    with open(findings_file_path) as f:
        d = json.load(f)
        r = d["results"]
        assert r[0]["check_id"] == "bash.curl.security.curl-eval.curl-eval"
        m = r[0]["extra"]["message"]
        assert (
            m
            == "Data is being eval'd from a `curl` command. An attacker with control of the server in the `curl` command could inject malicious code into the `eval`, resulting in a system comrpomise. Avoid eval'ing untrusted data if you can. If you must do this, consider checking the SHA sum of the content returned by the server to verify its integrity."
        )
        print("CHECK CHILDREN")
        for k in r:
            with pytest.raises(AssertionError):
                check_children(
                    k,
                    excludes=[
                        "cwe",
                        "path",
                        "check_id",
                        "license",
                        "taint_sink",
                        "taint_source",
                        "fingerprint",
                    ],
                )
    assert saw_message is True


@patch("verinfast.user.__get_input__", return_value="y")
def test_truncate(self):
    global saw_message
    saw_message = False
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    config.output_dir = results_dir
    assert config.runGit is True
    assert config.runScan is True
    agent.config = config
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.config.truncate_findings = True
    agent.config.truncate_findings_length = 0
    assert agent.config.runGit is True
    assert agent.config.runScan is True
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.scan()
    findings_file_path = results_dir.joinpath("small-test-repo.git.findings.json")

    with open(findings_file_path) as f:
        d = json.load(f)
        r = d["results"]
        assert r[0]["check_id"] == "bash.curl.security.curl-eval.curl-eval"
        assert r[0]["extra"]["lines"] == ""
        for k in r:
            check_children(k)
    assert saw_message is True


@patch("verinfast.user.__get_input__", return_value="y")
def test_truncate_from_args(self):
    global saw_message
    saw_message = False
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    parser = config.init_argparse()
    args = parser.parse_args(["--truncate_findings=30"])
    assert args.truncate_findings == 30
    config.handle_args(args)
    assert config.runGit is True
    assert config.runScan is True
    assert config.truncate_findings is True
    assert config.truncate_findings_length == 30
    config.output_dir = results_dir
    agent.config = config
    assert agent.config.truncate_findings is True
    assert agent.config.truncate_findings_length == 30
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.scan()
    findings_file_path = results_dir.joinpath("small-test-repo.git.findings.json")

    with open(findings_file_path) as f:
        d = json.load(f)
        r = d["results"]
        assert r[0]["check_id"] == "bash.curl.security.curl-eval.curl-eval"
        for k in r:
            check_children(k)
    assert saw_message is True



#!/usr/bin/env python3
from datetime import date
import json
import os
from pathlib import Path
import platform
import re
import shutil
import subprocess
import traceback
from uuid import uuid4

from cachehash.main import Cache
from modernmetric.__main__ import main as modernmetric

import httpx
from jinja2 import Environment, FileSystemLoader, select_autoescape
from pygments_tsx.tsx import patch_pygments

from verinfast.utils.utils import (
    DebugLog,
    std_exec,
    trimLineBreaks,
    escapeChars,
    truncate,
    truncate_children,
    get_repo_name_url_and_branch,
)
from verinfast.system.sysinfo import get_system_info
from verinfast.upload import Uploader
from verinfast.cloud.aws.costs import get_aws_costs
from verinfast.cloud.aws.get_profile import find_profile
from verinfast.cloud.azure.costs import runAzure
from verinfast.cloud.aws.instances import get_instances as get_aws_instances
from verinfast.cloud.azure.instances import get_instances as get_az_instances
from verinfast.cloud.gcp.instances import get_instances as get_gcp_instances
from verinfast.cloud.aws.blocks import getBlocks as get_aws_blocks
from verinfast.cloud.azure.blocks import getBlocks as get_az_blocks
from verinfast.cloud.gcp.blocks import getBlocks as get_gcp_blocks
from verinfast.config import Config
from verinfast.user import initial_prompt, save_path, repeat_boolean_prompt
from verinfast.utils.license import report as report_license
from verinfast.code_scan import run_scan

from verinfast.dependencies.walk import walk as dependency_walk


patch_pygments()

today = date.today()

requestx = httpx.Client(http2=True, timeout=None)
uname = platform.uname()
system = uname.system
node = uname.node
release = uname.release
version = uname.version
machine = uname.machine


template_definition = {}

file_path = Path(__file__)
parent_folder = file_path.parent.absolute()
templates_folder = str(parent_folder.joinpath("templates"))
# str_path = str(parent_folder.joinpath('str_conf.yaml').absolute())

curr_dir = os.getcwd()
temp_dir = Path(os.path.expanduser("~/.verinfast/")).joinpath("temp_repo")


class Agent:
    def __init__(self):
        self.config = Config()
        os.makedirs(self.config.output_dir, exist_ok=True)
        self.debug = DebugLog(file=self.config.log_file, debug=False)
        self.log = self.debug.log
        self.log(msg="", tag="Started")
        self.uploader = Uploader(self.config.upload_conf)
        self.up = self.uploader.make_upload_path
        self.config.upload_logs = initial_prompt()
        self.directory = save_path()

        # Initialize cache
        cache_dir = Path(Path.home(), ".verinfast_cache")
        db_path = Path(cache_dir, "semgrep.db")
        if not db_path.parent.exists():
            db_path.parent.mkdir(parents=True, exist_ok=True)

        self.cache = Cache(db_path, "semgrep_cache")

    def create_template(self):
        if not self.config.dry:
            with open(f"{self.config.output_dir}/results.html", "w") as f:
                jinja_env = Environment(
                    loader=FileSystemLoader(templates_folder),
                    autoescape=select_autoescape(["html", "xml"]),
                )
                jinja_env.globals.update(zip=zip, sorted=sorted)
                output = jinja_env.get_template("results.j2").render(
                    template_definition
                )
                f.write(output)

    def scan(self):
        report_license(self.config.reportId, self.config, "verinfast")
        if self.config.modules is not None:
            self.system_info = get_system_info()

            self.log(
                msg=json.dumps(self.system_info, indent=4), tag="System Information"
            )
            if not self.config.dry:
                system_info_file = os.path.join(
                    self.config.output_dir, "system_info.json"
                )
                try:
                    with open(system_info_file, "w") as f:
                        json.dump(self.system_info, f, indent=4)
                except IOError as e:
                    self.log(
                        f"Failed to write system info to {system_info_file}: {str(e)}"
                    )
                    raise RuntimeError(
                        f"Failed to write system information: {str(e)}"
                    ) from e

            if self.config.modules.code is not None:

                if self.config.shouldUpload:
                    headers = {
                        "content-type": "application/json",
                        "Accept-Charset": "UTF-8",
                    }
                    get_url = self.uploader.make_upload_path(
                        "scan_id", report=self.config.reportId
                    )
                    self.log(
                        msg=f"{self.config.baseUrl}{get_url}",
                        tag="Report Run Id Fetch",
                        display=True,
                    )
                    response = requestx.get(
                        f"{self.config.baseUrl}{get_url}", headers=headers
                    )
                    self.scanId = response.text.replace("'", "").replace('"', "")
                    if self.scanId and self.scanId != "":
                        self.log(msg=self.scanId, tag="Report Run Id", display=True)
                    else:
                        raise Exception(
                            f"{self.scanId} returned for failed report Id fetch."
                        )
                else:
                    self.log(
                        msg="Scan ID only fetched when uploading enabled",
                        tag="Scan ID",
                        display=True,
                    )
                self.scanRepos()
            if (
                self.config.modules
                and self.config.modules.cloud
                and len(self.config.modules.cloud)
            ):
                self.scanCloud()
            try:
                self.create_template()
            except:
                self.log(tag="ERROR", msg="Template Creation Failed")
        self.log(msg="", tag="Finished")

    # Excludes files in .git directories. Takes path of full path with filename
    def allowfile(self, path, allowDir=False):
        normpath = os.path.normpath(path)
        dirlist = normpath.split(os.sep)
        if (
            "node_modules" not in dirlist
            and ".git" not in dirlist
            and not os.path.islink(path)
            and (os.path.isfile(path) or allowDir)
        ):
            return True
        else:
            return False

    # Get recursive size of a directory
    def get_raw_size(self, start_path="."):
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(start_path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if os.path.isfile(fp) and not os.path.islink(fp):
                    total_size += os.path.getsize(fp)  # os.stat(fp).st_size
        return total_size

    # Simple algorithm for lines of code in a file
    def getloc(self, file):
        try:
            count = 0
            with open(file) as fp:
                for line in fp:
                    if line.strip():
                        count += 1
            return count
        except:
            return 0

    def checkDependency(self, command, name, kill=False) -> bool:
        which = shutil.which(command)
        if not which:
            self.log(
                msg=f"{name} is required but it's not installed.",
                tag=f"{name} status",
                display=False,
                timestamp=False,
            )
            if kill:
                raise Exception(f"{name} is required but it's not installed.")
            return False
        else:
            self.log(
                msg=f"{name} is installed at {which}.",
                tag=f"{name} status",
                display=True,
                timestamp=False,
            )
            return True

    def upload(self, file: str, route: str, source: str = "", isJSON=True):
        if not self.config.shouldUpload:
            self.log(
                msg="Skipping Uploads",
                tag=f"Skipping uploading {file} for {source} to {self.config.baseUrl}/{route}.",
                display=True,
            )
            return True
        if not Path(file).exists():
            self.log(msg=f"File does not exist: {file}")
            return False

        orig_route = route

        route = self.up(
            path_type=route,
            report=self.config.reportId,
            code=self.scanId,
            repo_name=source,
        )

        with open(file, "rb") as f:
            self.log(msg=f"{self.config.baseUrl}{route}", tag="Uploading to")

            if isJSON:
                headers = {
                    "Content-Type": "application/json",
                    "accept": "application/json",
                }
                response = requestx.post(
                    self.config.baseUrl + route, data=f, headers=headers
                )
            else:
                files = {"logFile": f}
                response = requestx.post(self.config.baseUrl + route, files=files)

        if response.status_code == 200:
            self.log(
                msg="",
                tag=f"Successfully uploaded {file} for {source} to {self.config.baseUrl}{route}.",
                display=True,
            )
            try:
                err_path_str = file[0:-5] + ".err"
                err_path = Path(err_path_str)

                if err_path.exists():
                    self.upload(
                        file=err_path_str,
                        route="err_" + orig_route,
                        source=source + " Error Logs",
                        isJSON=False,
                    )
            except:
                pass
            return True
        else:
            self.log(
                msg=response.status_code,
                tag=f"Failed to upload {file} for {source} to {self.config.baseUrl}{route}",
                display=True,
            )
            return False

    def formatGitHash(self, hash: str):
        hash = hash.replace("'", "").replace('"', "")
        message = std_exec(["git", "log", "-n1", "--pretty=format:%B", hash], self.log)
        author = std_exec(
            ["git", "log", "-n1", "--pretty=format:%aN <%aE>", hash], self.log
        )
        commit = std_exec(["git", "log", "-n1", "--pretty=format:%H", hash], self.log)
        date = std_exec(["git", "log", "-n1", "--pretty=format:%aD", hash], self.log)
        signed = std_exec(["git", "show", "--format='%G?'", hash], self.log)
        if signed != "N":
            signed = True
        else:
            signed = False
        merge = False
        merge1 = std_exec(["git", "show", hash], self.log)
        if merge1.startswith("Merge: "):
            merge = True
        returnVal = {
            "message": trimLineBreaks(message),
            "author": author,
            "commit": commit,
            "date": escapeChars(date),
            "signed": signed,
            "merge": merge,
        }
        return returnVal

    def parseRepo(self, path: str, repo_name: str, branch: str = None):
        self.log(msg="parseRepo")

        if not self.config.dry:
            os.chdir(path)

        if branch is None:
            branch = "main"

        # Adding this for Windows support
        # Appears to fail with blank HEAD
        try:
            std_exec(["git", "init"])
        except subprocess.CalledProcessError:
            self.log(
                msg="Error initializing git repository. This may not be a git repository.",
                tag="Git Init Error",
                display=True,
            )
            return

        git_dir_exists = os.path.isdir(os.path.join(path, ".git"))
        git_output_file = os.path.join(
            self.config.output_dir, repo_name + ".git.log.json"
        )

        if self.config.runGit and self.checkDependency("git", "Git") and git_dir_exists:
            try:
                if not self.config.dry:
                    subprocess.check_call(
                        ["git", f"--work-tree={path}", "checkout", branch]
                    )
            except subprocess.CalledProcessError:
                try:
                    if not self.config.dry:
                        subprocess.check_call(["git", "checkout", "master"])
                        branch = "master"
                except subprocess.CalledProcessError:
                    try:
                        cmd = "git for-each-ref --count=1 --sort=-committerdate refs/heads/ --format='%(refname:short)'"
                        # remove new lines and apostrophes from branch name.
                        branch = (
                            std_exec(cmd.split(" ")).replace("'", "").replace("\n", "")
                        )
                        subprocess.check_call(["git", "checkout", branch])
                    except subprocess.CalledProcessError:
                        self.log("Error checking out branch from git.", display=True)
            self.log(
                msg=repo_name, tag="Gathering source code statistics for", display=True
            )
            command = f"""git log \
                --since="{self.config.modules.code.git.start}" \
                --numstat \
                --format='%H' \
                {branch} --
            """
            gitlog = None
            try:
                if not self.config.dry:
                    results = subprocess.run(
                        command, shell=True, stdout=subprocess.PIPE
                    )
                    gitlog = results.stdout.decode()
            except subprocess.CalledProcessError:
                self.log("Error getting log from git", display=True)

            if not self.config.dry and gitlog:
                resultArr = gitlog.split("\n")
                prevHash = ""
                filesArr = []
                finalArr = []

                for line in resultArr:
                    lineArr = line.split("\t")
                    if len(lineArr) > 1:
                        filesArr.append(
                            {
                                "insertions": lineArr[0],
                                "deletions": lineArr[1],
                                "path": lineArr[2],
                            }
                        )
                    else:
                        if len(lineArr) == 1 and lineArr[0] != "":
                            # Hit next file
                            if prevHash:
                                # Not first one
                                hashObj = self.formatGitHash(prevHash)
                                hashObj["paths"] = filesArr
                                finalArr.append(hashObj)
                                filesArr = []
                            prevHash = lineArr[0]

                self.log(msg=truncate(finalArr), tag=f"{repo_name} Git Stats")

                self.log(msg=git_output_file, display=True)

                with open(git_output_file, "w") as f:
                    f.write(json.dumps(finalArr, indent=4))
                template_definition["gitlog"] = finalArr
                # End if not self.config.dry:
        else:
            self.log(
                msg="No git found. Check if the repository is a git repository if not expected.",
                tag=".git Missing",
                display=True,
            )

        # if Path(git_output_file).exists():
        self.upload(file=git_output_file, route="git", source=repo_name)

        # Manual File Sizes and Info
        sizes_output_file = os.path.join(
            self.config.output_dir, repo_name + ".sizes.json"
        )

        if not self.config.dry:
            self.log(msg=repo_name, tag="Gathering file sizes for", display=True)
            # Sizes for writing to output file
            # Intialize file list with "." as total size
            repo_size = self.get_raw_size(".")
            git_size = self.get_raw_size("./.git")

            # get sizes
            real_size = repo_size - git_size
            self.log(msg=repo_size, tag="Repo Size")
            self.log(msg=git_size, tag="Git Size")
            sizes = {
                "files": {
                    ".": {"size": repo_size, "loc": 0, "ext": None, "directory": True}
                },
                "metadata": {
                    "env": machine,
                    "real_size": real_size,
                    "uname": system,
                    "branch": locals()["branch"] if "branch" in locals() else None,
                },
            }

            # filelist for modernmetric
            filelist = []

            for filepath, subdirs, list in os.walk("."):
                for name in list:
                    fp = os.path.join(filepath, name)
                    extRe = re.search(r"^[^\.]*\.(.*)", name)
                    ext = extRe.group(1) if extRe else ""
                    if self.allowfile(path=fp):
                        if self.config.shouldManualFileScan:
                            file = {
                                "size": os.path.getsize(fp),
                                "loc": self.getloc(fp),
                                "ext": ext,  # os.path.splitext(name)[1],
                                "directory": False,
                            }
                            sizes["files"][fp] = file
                        filelist.append({"name": name, "path": fp})

            with open(sizes_output_file, "w") as f:
                f.write(json.dumps(sizes, indent=4))
            template_definition["current_dir_size"] = sizes["files"].pop(".")
            template_definition["sizes"] = sizes
            # End if not self.config.dry:

        self.upload(file=sizes_output_file, route="sizes", source=repo_name)

        # Run Modernmetric
        if self.config.runStats:
            stats_input_file = os.path.join(
                self.config.output_dir, repo_name + ".filelist.json"
            )
            stats_output_file = os.path.join(
                self.config.output_dir, repo_name + ".stats.json"
            )

            if not self.config.dry:
                self.log(
                    msg=repo_name,
                    tag="Analyzing repository with Modernmetric",
                    display=True,
                )
                with open(stats_input_file, "w") as f:
                    f.write(json.dumps(filelist, indent=4))

                template_definition["filelist"] = filelist

                custom_args = [
                    f"--file={stats_input_file}",
                    f"--output={stats_output_file}",
                ]
                modernmetric(
                    custom_args=custom_args, license_identifier=self.config.reportId
                )
                report_license(self.config.reportId, self.config, "modernmetric")

                with open(stats_output_file, "r") as f:
                    template_definition["stats"] = json.load(f)
            self.upload(file=stats_output_file, route="stats", source=repo_name)

        # Run SEMGrep
        run_scan(
            path,
            repo_name,
            self.config,
            self.cache,
            self.upload,
            template_definition,
            self.log,
        )

        # ##### Scan Dependencies ######
        if self.config.runDependencies:
            dependencies_output_file = os.path.join(
                self.config.output_dir, f"{repo_name}.dependencies.json"
            )
            self.log(msg=repo_name, tag="Scanning dependencies", display=True)
            if not self.config.dry:
                dependencies_output_file = dependency_walk(
                    output_file=dependencies_output_file, logger=self.log
                )
                with open(dependencies_output_file, "r") as f:
                    template_definition["dependencies"] = json.load(f)
            self.log(msg=dependencies_output_file, tag="Dependency File", display=False)
            self.upload(
                file=dependencies_output_file, route="dependencies", source=repo_name
            )

    def preflight(self):
        # If the 'dry' configuration is set, skip the preflight checks
        if self.config.dry:
            return
        # Loop over all remote repositories from config file
        print(
            "\n\n\nChecking your system's compatibility with the scan "
            "configuration:\n"
        )
        if "repos" in self.config.config:
            repos = self.config.config["repos"]
            if repos:
                # ignore blank lines from server
                for repo_url in [r for r in repos if len(r) > 0]:
                    match = re.search(r"([^/]*\.git.*)", repo_url)
                    if match:
                        repo_name = match.group(1)
                    else:
                        repo_name = repo_url.rsplit("/", 1)[-1]
                    if "@" in repo_name and re.search(r"^.*@.*\..*:", repo_url):
                        repo_url = "@".join(repo_url.split("@")[0:2])
                    elif "@" in repo_name:
                        repo_url = repo_url.split("@")[0]
                    try:
                        subprocess.check_output(["git", "ls-remote", repo_url])
                        self.log(
                            tag="Repository access confirmed",
                            msg=repo_url,
                            display=True,
                            timestamp=False,
                        )
                    except subprocess.CalledProcessError:
                        self.log(
                            msg=repo_url,
                            tag="Unable to access",
                            display=True,
                            timestamp=False,
                        )
                        self.log(
                            msg=repo_url,
                            tag="Repository will not be scanned",
                            display=True,
                            timestamp=False,
                        )

        cloud_config = self.config.modules.cloud
        if cloud_config is not None:
            for provider in cloud_config:
                try:
                    if provider.provider == "aws" and self.checkDependency(
                        "aws", "AWS Command-line tool"
                    ):
                        account_id = str(provider.account).replace("-", "")
                        if find_profile(account_id, self.log) is None:
                            self.log(
                                tag=f"No matching AWS CLI profiles found for {provider.account}",
                                msg="Account can't be scanned.",
                                display=True,
                                timestamp=False,
                            )
                        else:
                            self.log(
                                tag="AWS account access confirmed",
                                msg=account_id,
                                display=True,
                                timestamp=False,
                            )
                    if provider.provider == "azure" and self.checkDependency(
                        "az", "Azure Command-line tool"
                    ):
                        pass
                    if provider.provider == "gcp" and self.checkDependency(
                        "gcloud", "Google Command-line tool"
                    ):
                        pass
                except:
                    self.log(
                        msg=f"Unable to access {provider.provider} {provider.account}",
                        tag="Unable to access",
                        display=True,
                        timestamp=False,
                    )

        resp = repeat_boolean_prompt(
            "\nWould you like to proceed with the scan?", logger=print, default_val=True
        )

        if resp:
            self.log(msg="Proceeding")
        else:
            self.log(tag="Exiting now", msg="", display=True)
            exit(0)

    # ##### Scan Repos ######
    def scanRepos(self):
        # Loop over all remote repositories from config file
        if "repos" in self.config.config:
            repos = self.config.config["repos"]
            if repos:
                for repo_url in [
                    r for r in repos if len(r) > 0
                ]:  # ignore blank lines from server
                    repo_info = get_repo_name_url_and_branch(repo_url)
                    repo_name = repo_info["repo_name"]
                    repo_url = repo_info["repo_url"]
                    branch = repo_info["branch"]
                    self.log(msg=repo_name, tag="Processing", display=True)
                    self.log(msg=repo_url, tag="URL", display=True)
                    self.log(msg=branch, tag="Branch Specified", display=True)
                    try:
                        if not self.config.dry:
                            os.makedirs(temp_dir)
                    except:
                        self.log(tag="Directory exists:", msg=temp_dir, display=True)
                        try:
                            shutil.rmtree(temp_dir)
                            os.makedirs(temp_dir)
                        except Exception as e:
                            self.log(
                                tag=f"Failed to delete {temp_dir}", msg=e, display=True
                            )
                            continue

                    self.log(msg=repo_url, tag="Repo URL")
                    self.log(msg=temp_dir, tag="Temp Directory")
                    if not self.config.dry and self.config.runGit:
                        try:
                            subprocess.check_output(
                                ["git", "clone", repo_url, temp_dir]
                            )
                        except subprocess.CalledProcessError:
                            self.log(msg=repo_url, tag="Failed to clone", display=True)
                            continue

                        self.log(msg=repo_url, tag="Successfully cloned", display=True)
                    try:
                        self.parseRepo(temp_dir, repo_name, branch)
                    except Exception as e:
                        self.log(msg=str(e), tag="parseRepo Error Caught")
                        self.log(tag="", msg=traceback.format_exc())

                    os.chdir(curr_dir)
                    if not self.config.dry and self.config.delete_temp:
                        shutil.rmtree(temp_dir)
            else:
                self.log(msg="", tag="No remote repos", display=True)

        else:
            self.log(msg="", tag="No remote repos", display=True)

        # Loop over all local repositories from config file
        if "local_repos" in self.config.config:
            localrepos = self.config.config["local_repos"]
            if localrepos:
                for repo_path in localrepos:

                    # Check for a branch
                    split_url = repo_path.split("@")
                    branch = None
                    if len(split_url) == 2:
                        branch = split_url[1]
                        repo_path = split_url[0]

                    a = Path(repo_path).absolute()
                    match = re.search(r"([^/]*\.git.*)", str(a))
                    if match:
                        repo_name = match.group(1)
                    else:
                        repo_name = os.path.basename(os.path.normpath(repo_path))
                    self.parseRepo(repo_path, repo_name, branch)
            else:
                self.log(msg="", tag="No local repos", display=True)
        else:
            self.log(msg="", tag="No local repos", display=True)

        self.log(msg="", tag="Finished repo scans")

    # ##### Scan Cloud ######
    def scanCloud(self):
        self.log(msg="", tag="Doing cloud scan", display=True)
        cloud_config = self.config.modules.cloud
        self.log(msg=cloud_config, tag="Cloud Config")

        if cloud_config is None:
            return
        for provider in cloud_config:
            try:
                # Check if AWS-CLI is installed
                if provider.provider == "aws" and self.checkDependency(
                    "aws", "AWS Command-line tool"
                ):
                    account_id = str(provider.account).replace("-", "")
                    aws_cost_file = get_aws_costs(
                        targeted_account=account_id,
                        start=provider.start,
                        end=provider.end,
                        profile=provider.profile,
                        path_to_output=self.config.output_dir,
                        log=self.log,
                        dry=self.config.dry,
                    )
                    if aws_cost_file is None:
                        self.log(msg="Error processing AWS costs", tag=account_id)
                    else:
                        self.log(msg=aws_cost_file, tag="AWS Costs")
                        self.upload(file=aws_cost_file, route="costs", source="AWS")
                    aws_instance_file = get_aws_instances(
                        sub_id=account_id,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry,
                    )
                    if aws_instance_file is None:
                        self.log(msg="Error processing AWS instances", tag=account_id)
                    else:
                        self.log(msg=aws_instance_file, tag="AWS Instances")
                        self.upload(
                            file=aws_instance_file, route="instances", source="AWS"
                        )
                    aws_utilization_file = os.path.join(
                        self.config.output_dir,
                        f"aws-instances-{account_id}-utilization.json",
                    )
                    if Path(aws_utilization_file).is_file():
                        self.upload(
                            file=aws_utilization_file, route="utilization", source="AWS"
                        )
                    aws_block_file = get_aws_blocks(
                        sub_id=account_id,
                        path_to_output=self.config.output_dir,
                        log=self.log,
                        dry=self.config.dry,
                    )
                    if aws_block_file is None:
                        self.log(msg="Error processing AWS blocks", tag=account_id)
                    else:
                        self.log(msg=aws_block_file, tag="AWS Storage")
                        self.upload(file=aws_block_file, route="storage", source="AWS")

                # Check if Azure CLI is installed
                if provider.provider == "azure" and self.checkDependency(
                    "az", "Azure Command-line tool"
                ):
                    azure_cost_file = runAzure(
                        subscription_id=provider.account,
                        start=provider.start,
                        end=provider.end,
                        path_to_output=self.config.output_dir,
                        log=self.log,
                        dry=self.config.dry,
                    )
                    if azure_cost_file is None:
                        self.log(
                            msg="Error processing Azure costs", tag=provider.account
                        )
                    else:
                        self.log(msg=azure_cost_file, tag="Azure Costs")
                        self.upload(file=azure_cost_file, route="costs", source="Azure")
                    azure_instance_file = get_az_instances(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry,
                        log=self.log,
                    )
                    if azure_instance_file is None:
                        self.log(
                            msg="Error processing Azure instances", tag=provider.account
                        )
                    else:
                        self.log(msg=azure_instance_file, tag="Azure instances")
                        self.upload(
                            file=azure_instance_file, route="instances", source="Azure"
                        )
                    azure_utilization_file = os.path.join(
                        self.config.output_dir,
                        f"az-instances-{provider.account}-utilization.json",
                    )
                    if Path(azure_utilization_file).is_file():
                        self.upload(
                            file=azure_utilization_file,
                            route="utilization",
                            source="AWS",
                        )
                    azure_block_file = get_az_blocks(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry,
                    )
                    if azure_block_file is None:
                        self.log(
                            msg="Error processing Azure blocks", tag=provider.account
                        )
                    else:
                        self.log(msg=azure_block_file, tag="Azure Storage")
                        self.upload(
                            file=azure_block_file, route="storage", source="Azure"
                        )

                if provider.provider == "gcp" and self.checkDependency(
                    "gcloud", "Google Command-line tool"
                ):
                    gcp_instance_file = get_gcp_instances(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry,
                    )
                    if gcp_instance_file is None:
                        self.log(
                            msg="Error processing GCP instances", tag=provider.account
                        )
                    else:
                        self.log(msg=gcp_instance_file, tag="GCP instances")
                        self.upload(
                            file=gcp_instance_file, route="instances", source="GCP"
                        )
                    gcp_utilization_file = os.path.join(
                        self.config.output_dir,
                        f"gcp-instances-{provider.account}-utilization.json",
                    )
                    if Path(gcp_utilization_file).is_file():
                        self.upload(
                            file=gcp_utilization_file, route="utilization", source="AWS"
                        )
                    gcp_block_file = get_gcp_blocks(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry,
                    )
                    if gcp_block_file is None:
                        self.log(
                            msg="Error processing GCP blocks", tag=provider.account
                        )
                    else:
                        self.log(msg=gcp_block_file, tag="GCP Storage")
                        self.upload(file=gcp_block_file, route="storage", source="GCP")
            except Exception as e:
                self.log(tag="ERROR", msg="Error processing provider", display=True)
                self.log(tag="ERROR PROVIDER", msg=str(provider), display=True)
                self.log(tag="ERROR", msg=e, display=True)
                self.log(tag="ERROR STACK", msg=traceback.format_exc())


def main():
    agent = Agent()
    try:
        agent.preflight()
        agent.scan()
    except Exception as e:
        agent.log(msg=str(e), tag="Main Scan Error Caught")
        if agent.config.upload_logs:
            agent.upload(route="logs", file=agent.config.output_dir + "/log.txt")
        raise e

    # If user opts to upload logs, copy the log to the verinfast directory for future upload
    if agent.config.upload_logs:
        new_folder_name = str(today.year) + str(today.month) + str(today.day)
        d = agent.directory
        os.makedirs(f"{d}/{new_folder_name}/", exist_ok=True)
        new_file_name = str(uuid4()) + ".txt"
        fp = f"{d}/{new_folder_name}/{new_file_name}"
        shutil.copy2(agent.config.log_file, fp)
        print(
            f"""The log for this run was copied to:
            {d}/{new_folder_name}/{new_file_name}"""
        )

    # If user opts to upload results, upload and non-uploaded logs
    if agent.config.shouldUpload:
        log_list = os.listdir(agent.config.output_dir)
        log_list.sort()  # Upload current log last
        for file in log_list:
            if file.endswith("log.txt") and not file.startswith("u_"):
                file_path = os.path.join(agent.config.output_dir, file)
                agent.upload(route="logs", file=file_path, source=file, isJSON=False)
                new_path = os.path.join(agent.config.output_dir, "u_" + file)
                os.rename(file_path, new_path)

    # We only do this if you have a remote config but didn't upload
    if not agent.config.shouldUpload and agent.config.is_original_path_remote():
        print("\n\n\nIMPORTANT: To upload results from this location please run")
        print(
            f"verinfast -c {agent.config.cfg_path} -o {agent.config.output_dir} --should_upload --dry"
        )


if __name__ == "__main__":
    main()
# For test runs from commandline. Comment out before packaging. # main()



from verinfast.config import UploadConfig
from typing import Union


class Uploader:
    def __init__(self, config: UploadConfig):
        self.config = config

    def make_upload_path(
        self,
        path_type: str,
        report: Union[str, int],
        code: Union[str, int] = None,
        repo_name: str = None,
    ) -> str:
        """make_upload_path
        make_upload_path is a convenience function for abstracting
        the url paths from the scanning logic.

        Args:
            path_type (str) : specifies the scan action triggering the upload
            report (Union[str, int]): specifies the identifier the server uses for this scan
            code (Union[str, int]) :
        """

        code_sep = self.config.code_separator or ""

        cost_sep = self.config.cost_separator or ""

        paths = {
            "git": f"{report}{code_sep}/{code}/{repo_name}/git",
            "sizes": f"{report}{code_sep}/{code}/{repo_name}/sizes",
            "pygount": f"{report}{code_sep}/{code}/{repo_name}/pygount",
            "stats": f"{report}{code_sep}/{code}/{repo_name}/stats",
            "findings": f"{report}{code_sep}/{code}/{repo_name}/findings",
            "dependencies": f"{report}{code_sep}/{code}/{repo_name}/dependencies",
            "costs": f"{report}{cost_sep}/costs",
            "instances": f"{report}{cost_sep}/instances",
            "utilization": f"{report}{cost_sep}/instance_utilization",
            "storage": f"{report}{cost_sep}/storage",
            "scan_id": f"{report}{code_sep}",
            "logs": f"{report}/agent_logs",
            "err_stats": f"{report}/agent_err/stats_err",
            "err_findings": f"{report}/agent_err/findings_err",
        }

        if report is None:
            raise Exception("Invocation Error: must supply either a report ID or UUID")

        requires_code = ["git", "sizes", "pygount", "stats", "findings", "dependencies"]

        if path_type in requires_code and code is None:
            raise Exception(f"Invocation Error: type {path_type} requires code")

        if path_type in requires_code and repo_name is None:
            raise Exception(f"Invocation Error: type {path_type} requires repo_name")

        return_path = paths[path_type]

        # logs never gets uuid prefix
        if self.config.uuid and path_type != "logs":
            return_path = "uuid/" + return_path

        if self.config.prefix is not None:
            return_path = self.config.prefix + return_path

        return return_path



import datetime
import json
import os
from typing import List

from azure.identity import DefaultAzureCredential
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.monitor.query import MetricsQueryClient, MetricAggregationType


from verinfast.cloud.cloud_dataclass import (
    Utilization_Datapoint as Datapoint,
    Utilization_Datum as Datum,
)


metric_name = "Percentage CPU"
d = datetime.timedelta(
    days=30, seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0
)

metric_identifiers = {"cpu": metric_name}
aggregations = [
    MetricAggregationType.MINIMUM,
    MetricAggregationType.AVERAGE,
    MetricAggregationType.MAXIMUM,
]


def get_metrics_for_instance(
    metrics_client: MetricsQueryClient, instance_id: str, instance_name: str
) -> List[Datum]:
    print("Get Metrics for Instance")
    data = []
    try:
        o = metrics_client.query_resource(
            resource_uri=instance_id,
            interval=datetime.timedelta(hours=1),
            aggregations=aggregations,
            metric_names=[metric_name],
            timespan=d,
        )
        for dp in o.metrics[0].timeseries[0].data:
            n = Datapoint(Minimum=dp.minimum, Maximum=dp.maximum, Average=dp.average)
            datum = Datum(Timestamp=dp.timestamp.timestamp(), cpu=n)
            data.append(datum)

    except Exception as e:
        print(e)
        pass

    return data


def get_instances(sub_id: str, path_to_output: str = "./", dry=False, log=None):

    if not dry:
        my_instances = []
        metrics = []
        client = ComputeManagementClient(
            credential=DefaultAzureCredential(), subscription_id=sub_id
        )
        metrics_client = MetricsQueryClient(
            credential=DefaultAzureCredential(), subscription_id=sub_id
        )
        networkClient = NetworkManagementClient(
            credential=DefaultAzureCredential(), subscription_id=sub_id
        )

        res = client.virtual_machines.list_all()

        for vm in res:
            try:
                s = vm.storage_profile.os_disk.managed_disk.id
                tgt = s.split("resourceGroups/")[1]
                tgt = tgt.split("/")[0]
                name = vm.name
                location = vm.location
                hw = vm.hardware_profile.vm_size
                iv = client.virtual_machines.instance_view(
                    vm_name=name, resource_group_name=tgt
                )
                status = iv.statuses[-1]
                network_interface = vm.network_profile.network_interfaces[0]
                nic_name = network_interface.id.split("networkInterfaces/")[1]
                nics = networkClient.network_interface_ip_configurations.list(
                    resource_group_name=tgt, network_interface_name=nic_name
                )
                subnet = ""
                public_ip = "n/a"
                for nic in nics:
                    subnet = nic.subnet.name
                    vnet_name = nic.subnet.id.split("/")[-3]
                    if hasattr(nic, "public_ip_address") and hasattr(
                        nic.public_ip_address, "ip_address"
                    ):
                        public_ip_name = nic.public_ip_address.id.split("/")[-1]
                        pi = networkClient.public_ip_addresses.get(tgt, public_ip_name)
                        public_ip = pi.ip_address
                architecture = "x86_64"

                if "D2p" in hw or "E2p" in hw:
                    architecture = "Arm"

                my_instance = {
                    "id": vm.id,
                    "name": name,
                    "type": hw,
                    "state": status.display_status,
                    "zone": vm.zones[0] if vm.zones else None,
                    "region": location,
                    "subnet": subnet,
                    "architecture": architecture,
                    "publicIp": public_ip,
                    "vpc": vnet_name,
                }
            except Exception as e:
                if log:
                    log(tag="Azure Get Instance Error", msg=str(e))
                continue
            my_instances.append(my_instance)
            m = get_metrics_for_instance(
                metrics_client=metrics_client, instance_id=vm.id, instance_name=name
            )
            d = {"id": vm.id, "metrics": [metric.dict for metric in m]}
            metrics.append(d)
        upload = {
            "metadata": {"provider": "azure", "account": str(sub_id)},
            "data": my_instances,
        }
    # End dry block

    azure_output_file = os.path.join(path_to_output, f"az-instances-{sub_id}.json")
    if not dry:
        with open(azure_output_file, "w") as outfile:
            outfile.write(json.dumps(upload, indent=4))

        upload["data"] = metrics
        with open(azure_output_file[:-5] + "-utilization.json", "w") as outfile2:
            outfile2.write(json.dumps(upload, indent=4))
    return azure_output_file


# Test code
# print(get_instances(sub_id="80dc7a6b-df94-44be-a235-7e7ade335a3c"))



import json

# from unittest.mock import patch
import os
from pathlib import Path

# import shutil

# from verinfast.agent import Agent
# from verinfast.config import Config
# from verinfast.utils.utils import DebugLog
# import verinfast.user
from verinfast.dependencies.walk import walk
from verinfast.dependencies.walkers.package_lock import PackageWalker as PW

# from verinfast.dependencies.walkers.classes import Entry


file_path = Path(__file__)
test_folder = file_path.parent
fixtures_folder = test_folder.joinpath("fixtures")
walker_folder = fixtures_folder.joinpath("package_lock_walker")
results_dir = test_folder.joinpath("results").absolute()


def enabled_logger(flag=True):
    def silent(msg, **kwargs):
        return None

    def logger(msg, **kwargs):
        print(msg)
        print(kwargs)

    if flag:
        return logger
    else:
        return silent


def test_package_lock_exists():
    package_lock_file_path = walker_folder.joinpath("package-lock.json")
    assert package_lock_file_path.exists()


def test_indirect():
    folder_path = walker_folder.absolute()
    file_path = folder_path.joinpath("package-lock.json")

    output_file = test_folder.joinpath("dependencies1.json")
    assert file_path.exists(), "Manifest doesn't exist"
    assert not output_file.exists(), "Results file exists"

    output_path = walk(
        path=folder_path, output_file=output_file, logger=enabled_logger(False)
    )

    with open(output_path) as output_file:
        output = json.load(output_file)
        assert len(output) >= 1
        first_dep = output[0]
        assert first_dep["name"] == "@ampproject/remapping"
        assert first_dep["specifier"] == "2.2.1"

    os.remove(output_path)
    assert not Path(output_path).exists()
    return None


def test_direct():
    packageWalker = PW(
        manifest_type="json",
        logger=enabled_logger(False),
        root_dir=walker_folder,
        manifest_files=["package-lock.json"],
    )
    packageWalker.walk(path=walker_folder)
    assert packageWalker.entries[0].name == "@ampproject/remapping"



import json
import os
import subprocess
from verinfast.cloud.aws.get_profile import find_profile


def get_aws_costs(
    targeted_account, start, end, path_to_output, log, profile=None, dry=False
):

    def _get_costs_and_usage(profile: str, aws_output_file: str, next_token=None):
        base_cmd = f"""
            aws ce get-cost-and-usage \
            --time-period Start={start},End={end} \
            --granularity=DAILY \
            --metrics "BlendedCost" \
            --group-by Type=DIMENSION,Key=SERVICE \
            --profile="{profile}" \
            --output=json"""

        if next_token:
            cmd = f'{base_cmd} --next-page-token="{next_token}" | cat'
        else:
            cmd = f"{base_cmd} | cat"

        try:
            results = subprocess.run(
                cmd, shell=True, stdout=subprocess.PIPE, check=True
            )

        except subprocess.CalledProcessError:
            log(msg="Error getting data from AWS CLI get-cost-and-usage", tag="AWS CLI")
            raise Exception("Error getting aws cli data.")

        text = results.stdout.decode()

        if not text or not isinstance(text, str):
            log(msg="No data returned from AWS CLI get-cost-and-usage", tag="AWS CLI")
            return None

        return json.loads(text)

    def _process_results(obj):
        charges = []
        results_by_time = obj.get("ResultsByTime", [])
        for charge in results_by_time:
            if charge.get("Groups"):
                for group in charge["Groups"]:
                    newCharge = {
                        "Date": charge["TimePeriod"]["Start"],
                        "Group": group["Keys"][0],
                        "Cost": group["Metrics"]["BlendedCost"]["Amount"],
                        "Currency": group["Metrics"]["BlendedCost"]["Unit"],
                    }
                    charges.append(newCharge)
        return charges

    def _get_all_costs(profile: str, aws_output_file: str):
        all_charges = []
        next_token = None

        while True:
            # Get results for current page
            obj = _get_costs_and_usage(profile, aws_output_file, next_token)
            if not obj:
                break

            # Process and append results
            charges = _process_results(obj)
            all_charges.extend(charges)

            # Check for next page
            next_token = obj.get("NextPageToken")
            if not next_token:
                break

            log(
                msg=f"Fetching next page of results with token: "
                f"{next_token[:10]}...",
                tag="AWS CLI",
            )

        # Create final upload object
        upload = {
            "metadata": {"provider": "aws", "account": str(targeted_account)},
            "data": all_charges,
        }

        # Write results to file
        with open(aws_output_file, "w") as outfile:
            outfile.write(json.dumps(upload, indent=4))

        return aws_output_file

    if profile is None:
        profile = find_profile(targeted_account=targeted_account, log=log)

    if profile is None:
        log(msg="No matching profiles found", tag=targeted_account)
        return None

    output_file = os.path.join(path_to_output, f"aws-cost-{targeted_account}.json")

    if not dry:
        _get_all_costs(profile, output_file)

    return output_file



import json
from unittest.mock import patch
import os
from pathlib import Path
from verinfast.agent import Agent


@patch("verinfast.user.__get_input__", return_value="y")
def test_local_scan(self):
    file_path = Path(__file__)
    test_folder = file_path.parent
    stats_path = test_folder.joinpath("results").joinpath("tsx_test.stats.json")
    size_path = test_folder.joinpath("results").joinpath("tsx_test.sizes.json")
    findings_path = test_folder.joinpath("results").joinpath("tsx_test.findings.json")
    dependencies_path = test_folder.joinpath("results").joinpath(
        "tsx_test.dependencies.json"
    )
    tsx_path = test_folder.joinpath("tsx_test")

    agent = Agent()
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.config.runGit = True
    agent.config.output_dir = test_folder.joinpath("results")
    os.makedirs(agent.config.output_dir, exist_ok=True)
    agent.config.config["local_repos"] = [tsx_path.absolute().as_posix()]

    agent.scanRepos()

    with open(stats_path) as output_file:
        output = json.load(output_file)
        assert output is not None

    assert os.path.exists(size_path)
    assert os.path.exists(findings_path)
    assert os.path.exists(dependencies_path)

    return None


@patch("verinfast.user.__get_input__", return_value="y")
def test_local_scan_branch(self):
    file_path = Path(__file__)
    test_folder = file_path.parent
    stats_path = test_folder.joinpath("results").joinpath("tsx_test.stats.json")
    size_path = test_folder.joinpath("results").joinpath("tsx_test.sizes.json")
    findings_path = test_folder.joinpath("results").joinpath("tsx_test.findings.json")
    dependencies_path = test_folder.joinpath("results").joinpath(
        "tsx_test.dependencies.json"
    )
    tsx_path = test_folder.joinpath("tsx_test")

    agent = Agent()
    agent.config.dry = False
    agent.config.shouldUpload = False
    agent.config.runGit = True
    agent.config.output_dir = test_folder.joinpath("results")
    os.makedirs(agent.config.output_dir, exist_ok=True)
    agent.config.config["local_repos"] = [tsx_path.absolute().as_posix() + "@main"]

    agent.scanRepos()

    with open(stats_path) as output_file:
        output = json.load(output_file)
        assert output is not None

    assert os.path.exists(size_path)
    assert os.path.exists(findings_path)
    assert os.path.exists(dependencies_path)

    return None



import os
from unittest.mock import patch
from pathlib import Path
import shutil

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.utils.utils import DebugLog

file_path = Path(__file__)
test_folder = file_path.parent.absolute()
results_dir = test_folder.joinpath("results").absolute()
str_path = str(test_folder.joinpath("str_conf.yaml").absolute())


@patch("verinfast.user.__get_input__", return_value="y")
def test_no_config(self):
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    config.output_dir = results_dir
    print(agent.config.output_dir)
    agent.config = config
    assert agent.config.use_uuid is True
    agent.config.dry = True
    agent.config.shouldUpload = True
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.uploader.config = config.upload_conf
    assert agent.config.use_uuid is True, f"Expected True, config {agent.config}"
    get_url = agent.uploader.make_upload_path("scan_id", report=agent.config.reportId)
    assert (
        get_url == "/report/uuid/9a6e8696-f93a-4402-a64e-342ccb37592b/CorsisCode"
    ), get_url
    agent.scan()
    assert Path(results_dir).exists()
    # Make sure there are no .json results files
    results_path = Path(results_dir)
    assert results_path.exists()

    # Check if there are any JSON files
    json_files = list(results_path.glob("*.json"))
    assert not json_files, f"Found JSON files: {json_files}"

    # Since this test creates it's on DebubLog, the results
    # dir will have two log files. One is timestamped and is
    # the start of the regular log the the agent creates on start.
    # The other is the debug log created by this test, "log.txt"
    assert "/log.txt" in agent.debug.file
    with open(agent.debug.file) as f:
        logText = f.read()
    assert "Error" not in logText
    assert "File does not exist:" in logText
    # Since this test does not run a real scan, the debug log
    # will have errors about file uploads. Confirm it attempts
    # to upload all the files.
    upload_fail_prefix = "File does not exist: "
    upload_fail_prefix = upload_fail_prefix + str(results_dir) + "/"
    assert upload_fail_prefix + "small-test-repo.git.git.log.json" in logText
    assert upload_fail_prefix + "small-test-repo.git.sizes.json" in logText
    assert upload_fail_prefix + "small-test-repo.git.stats.json" in logText
    assert upload_fail_prefix + "small-test-repo.git.findings.json" in logText
    assert upload_fail_prefix + "small-test-repo.git.dependencies.json" in logText



from typing import Dict, Any
import platform
import psutil
import shutil


def get_system_info() -> Dict[str, Any]:
    """
    Collects system information and
    returns it as a JSON-compatible dictionary.
    """
    return {
        "OS": platform.system(),
        "OS Version": platform.version(),
        "OS Release": platform.release(),
        "Machine": platform.machine(),
        "Processor": platform.processor(),
        "Hostname": platform.node(),
        "CPU Cores": psutil.cpu_count(logical=False),
        "Logical CPUs": psutil.cpu_count(logical=True),
        "Total RAM": (f"{round(psutil.virtual_memory().total / (1024**3), 2)} GB"),
        "Available RAM": (
            f"{round(psutil.virtual_memory().available / (1024**3), 2)} GB"
        ),
        "Used RAM": f"{round(psutil.virtual_memory().used / (1024**3), 2)} GB",
        "Disk Size": (f"{round(shutil.disk_usage('/').total / (1024**3), 2)} GB"),
        "Disk Used": f"{round(shutil.disk_usage('/').used / (1024**3), 2)} GB",
        "Disk Free": f"{round(shutil.disk_usage('/').free / (1024**3), 2)} GB",
        "Python Version": platform.python_version(),
    }



from datetime import datetime, timedelta
import json
import os
from typing import List, Union

import boto3
import botocore

import verinfast.cloud.aws.regions as r
from verinfast.cloud.aws.get_profile import find_profile
from verinfast.cloud.cloud_dataclass import (
    Utilization_Datapoint as Datapoint,
    Utilization_Datum as Datum,
)

regions = r.regions


def get_metric_for_instance(
    metric: str,
    instance_id: str,
    session,
    region: str,
    namespace: str = "AWS/EC2",
    unit: str = "Percent",
):
    try:
        client = session.client("cloudwatch", region_name=region)
        response = client.get_metric_statistics(
            Namespace=namespace,
            MetricName=metric,
            Dimensions=[
                {"Name": "InstanceId", "Value": instance_id},
            ],
            StartTime=datetime.utcnow() - timedelta(days=30),
            EndTime=datetime.utcnow(),
            Period=60 * 60,
            Statistics=["Average", "Maximum", "Minimum"],
            Unit=unit,
        )
        return response
    except botocore.exceptions.ClientError:
        pass


def parse_multi(datapoint: Union[dict, List[dict]]) -> Datapoint:
    dp_sum: float = 0
    dp_count: int = 0
    dp_min: float = 0
    dp_max: float = 0

    my_datapoint = Datapoint()
    my_datapoint.Timestamp = datapoint["Timestamp"]
    if type(datapoint) is not list:
        datapoint = [datapoint]
    for entry in datapoint:
        if "Average" in entry:
            e = entry["Average"]
            dp_sum += e
            dp_count += 1
        if "Minimum" in entry:
            dp_min += entry["Minimum"]
        if "Maximum" in entry:
            dp_max += entry["Maximum"]
    dp_count = max(dp_count, 1)
    my_datapoint.Average = dp_sum / dp_count
    my_datapoint.Maximum = dp_max / dp_count
    my_datapoint.Minimum = dp_min / dp_count

    return my_datapoint


def get_instance_utilization(instance_id: str, session, region: str) -> List[Datum]:
    cpu_resp = get_metric_for_instance(
        metric="CPUUtilization", instance_id=instance_id, session=session, region=region
    )
    mem_resp = get_metric_for_instance(
        metric="mem_used_percent",
        instance_id=instance_id,
        namespace="CWAgent",
        session=session,
        region=region,
    )

    hdd_resp = get_metric_for_instance(
        metric="disk_used_percent",
        instance_id=instance_id,
        namespace="CWAgent",
        session=session,
        region=region,
    )

    cpu_stats: List[Datapoint] = []
    mem_stats: List[Datapoint] = []
    hdd_stats: List[Datapoint] = []

    # each instance may have more than 1 CPU
    for datapoint in cpu_resp["Datapoints"]:
        summary = parse_multi(datapoint)
        cpu_stats.append(summary)

    # memory and disk are not collected by default
    # each instance may have more than one disk
    if "Datapoints" in hdd_resp:
        for datapoint in hdd_resp["Datapoints"]:
            summary = parse_multi(datapoint)
            hdd_stats.append(summary)
    # memory
    if "Datapoints" in mem_resp:
        mem_stats = [Datapoint.From(i) for i in mem_resp["Datapoints"]]

    data = []
    m = max(len(cpu_stats), len(mem_stats), len(hdd_stats))
    if m == 0:
        return []
    else:
        for i in range(m):
            datum = Datum(Timestamp=cpu_stats[i].Timestamp)
            if i < len(cpu_stats):
                datum.cpu = cpu_stats[i]
            else:
                datum.cpu = None
            if i < len(mem_stats):
                datum.mem = mem_stats[i]
            else:
                datum.mem = None
            if i < len(hdd_stats):
                datum.hdd = hdd_stats[i]
            else:
                datum.hdd = None
            data.append(datum)
    return data


def get_instances(
    sub_id: str, profile: str = None, log=None, path_to_output: str = "./", dry=False
):
    if profile is None:
        profile = find_profile(targeted_account=sub_id, log=log)

    if profile is None:
        log(msg="No matching profiles found", tag=sub_id)
        return None

    if not dry:
        right_session = boto3.Session(profile_name=profile)
        my_instances = []
        metrics = []
        for region in regions:
            try:
                client = right_session.client("ec2", region_name=region)
                paginator = client.get_paginator("describe_instances")
                page_iterator = paginator.paginate()
                for page in page_iterator:
                    reservations = page["Reservations"]
                    for reservation in reservations:
                        instances = reservation["Instances"]
                        for instance in instances:
                            tags = instance.get("Tags")
                            if tags is None:
                                name = instance["InstanceId"]
                            else:
                                tags_with_name = [
                                    t["Value"] for t in tags if t["Key"] == "Name"
                                ]
                                if not tags_with_name:
                                    name = instance["InstanceId"]
                                else:
                                    name = tags_with_name[0]
                            m = get_instance_utilization(
                                instance_id=instance["InstanceId"],
                                session=right_session,
                                region=region,
                            )
                            d = {
                                "id": instance["InstanceId"],
                                "metrics": [metric.dict for metric in m],
                            }
                            metrics.append(d)
                            region = None
                            subnet_id = None
                            zone = None
                            if "SubnetId" in instance:
                                subnet_id = instance["SubnetId"]
                            if "Placement" in instance:
                                placement = instance["Placement"]
                                if "AvailabilityZone" in placement:
                                    zone = placement["AvailabilityZone"]
                                    region = zone[0:-1]
                            try:
                                result = {
                                    "id": instance["InstanceId"],
                                    "name": name or "n/a",
                                    "state": instance["State"]["Name"],
                                    "type": instance["InstanceType"],
                                    "zone": zone or "n/a",
                                    "region": region or "n/a",
                                    "subnet": (subnet_id or "n/a"),
                                    "architecture": instance["Architecture"],
                                }
                            except Exception as e:
                                if log:
                                    log(tag="AWS Get Instance Error", msg=str(e))
                                continue
                            if "VpcId" in instance:
                                result["vpc"] = instance["VpcId"]
                            else:
                                result["vpc"] = "n/a"
                            if "PublicIpAddress" in instance:
                                result["publicIp"] = instance["PublicIpAddress"]
                            else:
                                result["publicIp"] = "n/a"
                            if "NetworkInterfaces" in instance:
                                ni = instance["NetworkInterfaces"]
                                for interface in ni:
                                    if "Association" in interface:
                                        association = interface["Association"]
                                        if "PublicIp" in association:
                                            public_ip = association["PublicIp"]
                                            result["publicIp"] = public_ip
                            my_instances.append(result)
            except Exception as e:
                if log:
                    log(tag="AWS Get Instance Error - Region", msg=str(e))
                pass

        upload = {
            "metadata": {"provider": "aws", "account": str(sub_id)},
            "data": my_instances,
        }
    # End dry check

    aws_output_file = os.path.join(path_to_output, f"aws-instances-{sub_id}.json")

    if not dry:
        with open(aws_output_file, "w") as outfile:
            outfile.write(json.dumps(upload, indent=4))
        upload["data"] = metrics
        with open(aws_output_file[:-5] + "-utilization.json", "w") as outfile2:
            outfile2.write(json.dumps(upload, indent=4))

    return aws_output_file



from verinfast.system.sysinfo import get_system_info
import json


def test_system_info_comprehensive():
    """Comprehensive test of system information collection"""
    print("\nTesting system information collection:")

    # Get system information
    info = get_system_info()

    # Print formatted output for visual inspection
    print("\nCollected System Information:")
    print(json.dumps(info, indent=4))

    # Validate the information
    assert isinstance(info, dict), "System info should be a dictionary"
    assert len(info) > 0, "System info should not be empty"

    # Check for essential keys
    essential_keys = [
        "OS",
        "OS Version",
        "OS Release",
        "Machine",
        "Processor",
        "CPU Cores",
        "Total RAM",
        "Python Version",
    ]

    print("\nValidating essential system information:")
    for key in essential_keys:
        assert key in info, f"Missing essential key: {key}"
        print(f"✓ {key}: {info[key]}")

    # Print summary
    print(f"\nTotal fields collected: {len(info)}")
    print("All validations passed!")


if __name__ == "__main__":
    test_system_info_comprehensive()



import json
from unittest.mock import patch
import os
from pathlib import Path
from verinfast.agent import Agent
import verinfast.user


@patch("verinfast.user.__get_input__", return_value="y")
def test_scan(self):
    file_path = Path(__file__)
    test_folder = file_path.parent
    repo_name = "test_tsx"
    output_path = test_folder.joinpath("results").joinpath(repo_name + ".stats.json")
    tsx_path = test_folder.joinpath("tsx_test")

    agent = Agent()
    agent.config.runGit = False
    agent.config.dry = False
    agent.config.runPygount = False
    agent.config.runScan = False
    agent.config.runSizes = False
    agent.config.runStats = True
    agent.config.output_dir = test_folder.joinpath("results")
    os.makedirs(agent.config.output_dir, exist_ok=True)

    agent.parseRepo(path=tsx_path, repo_name=repo_name)

    with open(output_path) as output_file:
        output = json.load(output_file)
        assert output is not None

    return None


@patch("verinfast.user.__get_input__", return_value="y")
def test_home(self):
    assert verinfast.user.user_home is not None


@patch("verinfast.user.__get_input__", return_value="y")
def test_mock(self):
    assert verinfast.user.__get_input__() == "y"



from typing import List, Optional, Dict, Any
from modernmetric import process_diff_content
from verinfast.utils.utils import std_exec


class MetricArgs:

    def __init__(self):
        self.ignore_lexer_errors = True
        self.dump = False


def analyze_git_diff(
    commit_id: str, file_name: str, log=None
) -> Optional[Dict[str, Any]]:

    cmd: List[str] = ["git", "show", "-p", "--format=", commit_id, "--", file_name]

    try:
        # Get diff output using std_exec
        diff_output = std_exec(cmd, log=log)

        # We're skipping the first 4 lines of git diff output
        # because they only contain metadata
        diff_lines = diff_output.splitlines()[4:]

        # Separate added and deleted lines
        added_lines = []
        deleted_lines = []

        for line in diff_lines:
            # Skip diff header lines that start with @
            if line.startswith("@"):
                continue
            # Added lines start with +
            elif line.startswith("+"):
                added_lines.append(line[1:])
            # Deleted lines start with -
            elif line.startswith("-"):
                deleted_lines.append(line[1:])

        added_content = "\n".join(added_lines)
        deleted_content = "\n".join(deleted_lines)

        args = MetricArgs()

        added_results = process_diff_content(
            _content=added_content, _file=file_name, _args=args, _importer={}
        )

        deleted_results = process_diff_content(
            _content=deleted_content, _file=file_name, _args=args, _importer={}
        )

        # Combine metrics
        combined_metrics = {}
        for metric_name, added_value in added_results[0].items():
            deleted_value = deleted_results[0].get(metric_name, 0)
            combined_metrics[metric_name] = added_value + deleted_value

        return {
            "metrics": combined_metrics,
            "file": file_name,
            "language": added_results[2],  # same language 4 added and deleted
            "store": {"added": added_results[4], "deleted": deleted_results[4]},
        }

    except Exception as e:
        if log is not None:
            log(tag="git_metrics Error", msg=f"Error analyzing git diff: {str(e)}")
        else:
            print(f"Error analyzing git diff: {str(e)}")
        return None



import json
import os
import pytest
from unittest.mock import patch, MagicMock
from verinfast.cloud.aws.costs import get_aws_costs

# Mock AWS CLI responses
MOCK_RESPONSE_1 = {
    "ResultsByTime": [
        {
            "TimePeriod": {"Start": "2024-01-01"},
            "Groups": [
                {
                    "Keys": ["AWS Service 1"],
                    "Metrics": {"BlendedCost": {"Amount": "100.00", "Unit": "USD"}},
                }
            ],
        }
    ],
    "NextPageToken": "mock-token-123",
}

MOCK_RESPONSE_2 = {
    "ResultsByTime": [
        {
            "TimePeriod": {"Start": "2024-01-01"},
            "Groups": [
                {
                    "Keys": ["AWS Service 2"],
                    "Metrics": {"BlendedCost": {"Amount": "200.00", "Unit": "USD"}},
                }
            ],
        }
    ]
}


@pytest.fixture
def mock_subprocess_run():
    with patch("subprocess.run") as mock_run:
        # First call returns response with NextPageToken
        first_response = MagicMock()
        first_response.stdout.decode.return_value = json.dumps(MOCK_RESPONSE_1)

        # Second call returns response without NextPageToken
        second_response = MagicMock()
        second_response.stdout.decode.return_value = json.dumps(MOCK_RESPONSE_2)

        mock_run.side_effect = [first_response, second_response]
        yield mock_run


def test_aws_costs_pagination(mock_subprocess_run, tmp_path):
    # Mock logger
    mock_logger = MagicMock()

    # Run the function
    output_file = get_aws_costs(
        targeted_account="123456789",
        start="2024-01-01",
        end="2024-01-31",
        path_to_output=str(tmp_path),
        log=mock_logger,
        profile="test-profile",
    )

    # Verify subprocess.run was called twice (pagination)
    assert mock_subprocess_run.call_count == 2

    # Verify the second call included the NextPageToken
    second_call_args = mock_subprocess_run.call_args_list[1][0][0]
    assert '--next-page-token="mock-token-123"' in second_call_args

    # Verify the output file exists and contains combined data
    assert os.path.exists(output_file)

    with open(output_file, "r") as f:
        result = json.load(f)

    # Verify metadata
    assert result["metadata"]["provider"] == "aws"
    assert result["metadata"]["account"] == "123456789"

    # Verify combined data
    assert len(result["data"]) == 2
    assert result["data"][0]["Cost"] == "100.00"
    assert result["data"][1]["Cost"] == "200.00"
    assert result["data"][0]["Group"] == "AWS Service 1"
    assert result["data"][1]["Group"] == "AWS Service 2"


def test_aws_costs_no_pagination(mock_subprocess_run, tmp_path):
    # Override mock to return only one response without NextPageToken
    mock_subprocess_run.side_effect = None
    response = MagicMock()
    response.stdout.decode.return_value = json.dumps(MOCK_RESPONSE_2)
    mock_subprocess_run.return_value = response

    mock_logger = MagicMock()

    output_file = get_aws_costs(
        targeted_account="123456789",
        start="2024-01-01",
        end="2024-01-31",
        path_to_output=str(tmp_path),
        log=mock_logger,
        profile="test-profile",
    )

    # Verify subprocess.run was called only once (no pagination)
    assert mock_subprocess_run.call_count == 1

    # Verify the output file exists and contains correct data
    assert os.path.exists(output_file)

    with open(output_file, "r") as f:
        result = json.load(f)

    # Verify data
    assert len(result["data"]) == 1
    assert result["data"][0]["Cost"] == "200.00"
    assert result["data"][0]["Group"] == "AWS Service 2"



from datetime import datetime, timedelta

import json
import os

import boto3

import verinfast.cloud.aws.regions as r
from verinfast.cloud.aws.get_profile import find_profile

regions = r.regions


def getBlocks(
    sub_id: str, profile: str = None, log=None, path_to_output: str = "./", dry=False
):
    if profile is None:
        profile = find_profile(targeted_account=sub_id, log=log)

    if profile is None:
        log(msg="No matching profiles found", tag=sub_id)
        return None

    if not dry:
        right_session = boto3.Session(profile_name=profile)
        s3 = right_session.client("s3", region_name=regions[0])
        response = s3.list_buckets()
        known_buckets = {}
        for bucket in response["Buckets"]:
            bucket_name = bucket["Name"]
            resp = s3.get_bucket_location(Bucket=bucket_name)

            permissions = []
            public = False
            try:
                policy_status_resp = s3.get_bucket_policy_status(Bucket=bucket_name)
                public = policy_status_resp["PolicyStatus"]["IsPublic"]

                if "Policy" in policy_status_resp:
                    p_string = policy_status_resp["Policy"]
                    p_dict = json.loads(p_string)
                    statements = p_dict["Statement"]
                    # print(statements)
                    permissions = [json.dumps(s) for s in statements]
                    # print(s2)
            except s3.exceptions.from_code("NoSuchBucketPolicy"):
                log(msg=bucket_name, tag="No Bucket Policy for bucket")

            versioning = None
            try:
                versioning_response = s3.get_bucket_versioning(Bucket=bucket_name)
                if "Status" in versioning_response:
                    versioning = versioning_response["Status"]

            except s3.exceptions.from_code("NoSuchBucketStatus"):
                log(msg=bucket_name, tag="No Bucket Status for bucket")

            region = resp["LocationConstraint"]
            # print(region)
            if region:
                # print('Have region')
                cloudwatch = right_session.client("cloudwatch", region_name=region)
            else:
                cloudwatch = right_session.client("cloudwatch", region_name="us-east-1")

            response = cloudwatch.get_metric_statistics(
                Namespace="AWS/S3",
                MetricName="BucketSizeBytes",
                Dimensions=[
                    {"Name": "BucketName", "Value": bucket_name},
                    {"Name": "StorageType", "Value": "StandardStorage"},
                ],
                StartTime=datetime.now() - timedelta(days=2),
                EndTime=datetime.now(),
                Period=3600,
                Statistics=["Average"],
                Unit="Bytes",
            )
            if response["Datapoints"]:
                bucket_size_bytes = response["Datapoints"][-1]["Average"]
                known_buckets[bucket_name] = {
                    "name": bucket_name,
                    "size": int(bucket_size_bytes),
                    "retention": versioning,
                    "public": public,
                    "permissions": permissions,
                }
            else:
                pass
                # print(response)
            # except:
            #     pass

        my_buckets = list(known_buckets.values())
        upload = {
            "metadata": {"provider": "aws", "account": str(sub_id)},
            "data": my_buckets,
        }
    # End dry block

    aws_output_file = os.path.join(path_to_output, f"aws-storage-{sub_id}.json")

    if not dry:
        with open(aws_output_file, "w") as outfile:
            outfile.write(json.dumps(upload, indent=4))

    return aws_output_file


# Test Code
# getBlocks(sub_id='436708548746')



import json
import os
from pathlib import Path
from typing import List

from verinfast.dependencies.walkers.classes import Walker, Entry
from verinfast.utils.utils import std_exec


class ComposerWalker(Walker):
    dependency_tree = {}

    def initialize(self, root_path: str = "./"):
        root_path = str(Path(root_path).absolute())
        self.install_points: List[Path] = []
        for p in Path(root_path).rglob("**/*.*"):
            if p.name in self.manifest_files:
                self.install_points.append(p)
        for p in self.install_points:
            target_dir = Path(p).parent
            os.chdir(target_dir)
            try:
                res = std_exec(["composer", "install", "--no-dev", "--no-progress"])
                self.log(tag="", msg=res, timestamp=False)
            except Exception as error:
                self.log(tag="ERROR", msg=f"Error with composer install: {error}")
            finally:
                self.walk("./")
            for dep in self.dependency_tree:
                dependency = self.dependency_tree[dep]
                for v in dependency:
                    version = dependency[v]
                    self.entries.append(version)
            os.chdir(root_path)

    def parse(self, file: str, expand=False):
        try:
            with open(file) as f:
                cjson = json.load(f)
                if "license" in cjson and self.dependency_tree.get(cjson["name"]):
                    for version in self.dependency_tree[cjson["name"]]:
                        self.dependency_tree[cjson["name"]][version].license = cjson[
                            "license"
                        ]
                        if "description" in cjson:
                            self.dependency_tree[cjson["name"]][version].summary = (
                                cjson["description"]
                            )
                if "require" in cjson:
                    for k, v in cjson["require"].items():
                        if self.dependency_tree.get(k):
                            self.dependency_tree[k] = self.dependency_tree[k]
                        else:
                            self.dependency_tree[k] = {}
                        if self.dependency_tree[k].get(v):
                            self.dependency_tree[k][v] = self.dependency_tree[k][v]
                        else:
                            self.dependency_tree[k][v] = Entry(
                                name=k, source="composer", specifier=v
                            )
                self.log(self.dependency_tree)
        except Exception as error:
            self.log(f"error parsing {file}")
            try:
                self.log(json.load(file)["name"])
                self.log(json.load(file)["require"])
            except Exception as ex:
                raise ex
            self.log(error)



import json
from pathlib import Path
from typing import List

import defusedxml

from verinfast.dependencies.walkers.composer import ComposerWalker
from verinfast.dependencies.walkers.maven import MavenWalker
from verinfast.dependencies.walkers.npm import NodeWalker
from verinfast.dependencies.walkers.package_lock import PackageWalker
from verinfast.dependencies.walkers.gemwalker import GemWalker
from verinfast.dependencies.walkers.nuget import NuGetWalker, c_sharp_matches
from verinfast.dependencies.walkers.dockerwalker import DockerWalker
from verinfast.dependencies.walkers.python import PyWalker
from verinfast.dependencies.walkers.go import GoWalker
from verinfast.dependencies.walkers.classes import Entry

defusedxml.defuse_stdlib()

# Manifests we support
# should probably move this to a conf

# TODO: Gemfile parse
# Example: https://github.com/mastodon/mastodon/blob/main/Gemfile
ruby_matches = ["Gemfile"]

# TODO: Parse Cargo.toml
# Example: https://github.com/servo/servo/blob/master/components/style/Cargo.toml
rust_matches = ["Cargo.toml"]


def write_file(output_file: str, entries):
    with open(output_file, "w") as outfile:
        dicts = [entry.to_json() for entry in entries]
        outfile.write(json.dumps(dicts, indent=4))


# Finds all manifests we can process in the repo
# and stores their path in memory
def walk(logger, path: str = "./", output_file="./dependencies.json"):
    mavenWalker = MavenWalker(
        manifest_type="xml", manifest_files=["pom.xml"], logger=logger, root_dir=path
    )

    nodeWalker = NodeWalker(
        manifest_type="json",
        manifest_files=["package.json"],
        logger=logger,
        root_dir=path,
    )

    nugetWalker = NuGetWalker(
        manifest_type="xml",
        manifest_files=c_sharp_matches,
        logger=logger,
        root_dir=path,
    )

    py_walker = PyWalker(
        manifest_files=[
            "requirements.txt",
            "requirements-dev.txt",
            "pyproject.toml",
            "poetry.lock",
        ],
        manifest_type="txt",
        logger=logger,
        root_dir=path,
    )
    gem_walker = GemWalker(
        manifest_files=["gemfile", "Gemfile"],
        manifest_type="ruby",
        logger=logger,
        root_dir=path,
    )
    docker_walker = DockerWalker(
        manifest_files=["Dockerfile", "dockerfile", "docker-compose.yml"],
        manifest_type="Dockerfile",
        logger=logger,
        root_dir=path,
    )
    composer_walker = ComposerWalker(
        manifest_files=["composer.json"],
        manifest_type="json",
        logger=logger,
        root_dir=path,
    )
    go_walker = GoWalker(
        manifest_type="sum", manifest_files=["go.sum"], logger=logger, root_dir=path
    )
    path = str(Path(path).absolute())
    entries: List[Entry] = []
    composer_walker.initialize(root_path=path)
    entries += composer_walker.entries
    mavenWalker.walk(path=path)
    logger(msg="Dependency Scan 10%", display=True)
    entries += mavenWalker.entries
    write_file(output_file=output_file, entries=entries)
    nodeWalker.initialize(root_path=path)
    logger(msg="Dependency Scan 25%", display=True)
    if nodeWalker.entries:
        entries += nodeWalker.entries
    else:
        packageWalker = PackageWalker(
            manifest_type="json",
            logger=logger,
            root_dir=path,
            manifest_files=["package-lock.json"],
        )
        packageWalker.walk(path=path)
        entries += packageWalker.entries
    write_file(output_file=output_file, entries=entries)
    nugetWalker.initialize()
    logger(msg="Dependency Scan 40%", display=True)
    nugetWalker.walk(path=path)
    entries += nugetWalker.entries
    write_file(output_file=output_file, entries=entries)
    py_walker.walk(path=path)
    logger(msg="Dependency Scan 60%", display=True)
    entries += py_walker.entries
    write_file(output_file=output_file, entries=entries)
    gem_walker.walk(path=path)
    logger(msg="Dependency Scan 80%", display=True)
    entries += gem_walker.entries
    write_file(output_file=output_file, entries=entries)
    logger(msg="Dependency Scan 95%", display=True)
    docker_walker.walk(path=path, expand=False)
    entries += docker_walker.entries
    write_file(output_file=output_file, entries=entries)
    logger(msg="Dependency Scan 100%", display=True)
    go_walker.walk(path=path)
    entries += go_walker.entries
    write_file(output_file=output_file, entries=entries)
    return output_file



import glob
import json
import subprocess
from pathlib import Path
from typing import List
import httpx
import os


class Entry(dict):
    def __init__(
        self,
        name: str,
        source: str,
        specifier: str = None,
        license: str = None,
        summary: str = None,
        requires=None,
        required_by=None,
    ) -> None:
        if not name:
            raise Exception("Entries must have a package name")
        if not source:
            raise Exception("Entries must have a package source")

        self.name = name
        self.source = source
        self.specifier = specifier
        self.license = license
        self.requires = requires
        self.required_by = required_by
        self.summary = summary

    def __str__(self) -> str:
        d = self.to_json()
        return json.dumps(d, indent=4)

    def __repr__(self) -> str:
        return str(self)

    def to_json(self) -> dict:
        d = {}
        d["name"] = self.name
        d["source"] = self.source
        if self.specifier:
            d["specifier"] = self.specifier
        if self.license:
            d["license"] = self.license
        if self.summary:
            d["summary"] = self.summary
        if self.required_by:
            d["required_by"] = self.required_by
        return d


class Walker:
    def __init__(
        self,
        manifest_type: str,  # "json",
        manifest_files: List[str],  # ["package.json"]
        logger,
        root_dir: str = "./",
        print_name: str = None,
    ) -> None:
        if print_name:
            logger(print_name)
        self.files = []
        new_files = []
        for f in manifest_files:
            if "*" in f:
                full_paths = glob.glob(os.path.join(root_dir, f))
                expanded = [os.path.basename(p) for p in full_paths]
                new_files.extend(expanded)
            else:
                new_files.append(f)
        self.manifest_files = new_files.copy()
        self.manifest_type = manifest_type
        self.entries = []
        self.requestx = httpx.Client(http2=True, timeout=None)
        self.loggerFunc = logger

    def initialize(self, command: str):
        if command:
            subprocess.call(args=command)

    def log(self, msg, tag=None, display=False, timestamp=True):
        self.loggerFunc(msg, tag=tag, display=display, timestamp=timestamp)

    def getUrl(self, url: str, headers: dict = {}):
        try:
            return self.requestx.get(url=url, headers=headers).content.decode(
                "utf-8-sig"
            )
        except Exception as e:
            self.loggerFunc(f"Failed to get URL: {url}, {e}", display=False)
            return None

    def walk(
        self, path: str = "./", parse: bool = True, expand: bool = False, debug: int = 0
    ):
        for p in Path(path).rglob("**/*"):
            if debug > 1:
                self.log(f"EVALUATING: {p}", display=(debug > 2))
            if p.name in self.manifest_files:
                if debug > 0:
                    self.log(f"FOUND: {p.name}", display=True)
                self.files.append(p)
                if parse:
                    self.parse(file=str(p.absolute()), expand=expand)

    def parse(self, file: str, expand=False):
        raise Exception("No parser for this Walker")

    def expand(self, file):
        raise Exception("No expansion for this Walker")
